@@ -3,7 +3,6 @@ from __future__ import print_function
 
 import argparse
 import configparser
-import dbus
 import glob
 import gzip
 import os
@@ -11,17 +10,19 @@ import re
 import struct
 import subprocess
 import sys
-
 from collections import defaultdict
-from dbus.mainloop.glib import DBusGMainLoop
+from datetime import datetime
 from errno import EACCES, EPERM
-from gi.repository import GLib
-from mmio import MMIO, MMIOError
 from multiprocessing import cpu_count
 from platform import uname
 from threading import Event, Thread
 from time import time
 
+import dbus
+from dbus.mainloop.glib import DBusGMainLoop
+from gi.repository import GLib
+from mmio import MMIO, MMIOError
+
 DEFAULT_SYSFS_POWER_PATH = '/sys/class/power_supply/AC*/online'
 VOLTAGE_PLANES = {'CORE': 0, 'GPU': 1, 'CACHE': 2, 'UNCORE': 3, 'ANALOGIO': 4}
 CURRENT_PLANES = {'CORE': 0, 'GPU': 1, 'CACHE': 2}
@@ -29,6 +30,9 @@ TRIP_TEMP_RANGE = [40, 97]
 UNDERVOLT_KEYS = ('UNDERVOLT', 'UNDERVOLT.AC', 'UNDERVOLT.BATTERY')
 ICCMAX_KEYS = ('ICCMAX', 'ICCMAX.AC', 'ICCMAX.BATTERY')
 power = {'source': None, 'method': 'polling'}
+HWP_VALUE = 0x20
+HWP_INTERVAL = 60
+
 
 platform_info_bits = {
     'maximum_non_turbo_ratio': [8, 15],
@@ -73,6 +77,7 @@ supported_cpus = {
     'Kaby Lake (R)': (0x8E, 0x9E),
     'Coffee Lake': (0x9E,),
     'Cannon Lake': (0x66,),
+    'Comet Lake': (0xA6,),
 }
 
 
@@ -88,14 +93,33 @@ OK = bcolors.GREEN + bcolors.BOLD + 'OK' + bcolors.RESET
 ERR = bcolors.RED + bcolors.BOLD + 'ERR' + bcolors.RESET
 LIM = bcolors.YELLOW + bcolors.BOLD + 'LIM' + bcolors.RESET
 
+log_history = set()
 
-def fatal(msg, code=1):
-    print('[E] {:s}'.format(msg), file=sys.stderr)
+
+def log(msg, oneshot=False, end='\n'):
+    outfile = args.log if args.log else sys.stdout
+    if msg.strip() not in log_history or oneshot is False:
+        tstamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
+        full_msg = '{:s}: {:s}'.format(tstamp, msg) if args.log else msg
+        print(full_msg, file=outfile, end=end)
+        log_history.add(msg.strip())
+
+
+def fatal(msg, code=1, end='\n'):
+    outfile = args.log if args.log else sys.stderr
+    tstamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
+    full_msg = '{:s}: [E] {:s}'.format(tstamp, msg) if args.log else '[E] {:s}'.format(msg)
+    print(full_msg, file=outfile, end=end)
     sys.exit(code)
 
 
-def warning(msg):
-    print('[W] {:s}'.format(msg), file=sys.stderr)
+def warning(msg, oneshot=True, end='\n'):
+    outfile = args.log if args.log else sys.stderr
+    if msg.strip() not in log_history or oneshot is False:
+        tstamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
+        full_msg = '{:s}: [W] {:s}'.format(tstamp, msg) if args.log else '[W] {:s}'.format(msg)
+        print(full_msg, file=outfile, end=end)
+        log_history.add(msg.strip())
 
 
 def writemsr(msr, val):
@@ -150,21 +174,6 @@ def readmsr(msr, from_bit=0, to_bit=63, cpu=None, flatten=False):
             raise e
 
 
-def cpu_usage_pct(exit_event, interval=1.0):
-    last_idle = last_total = 0
-
-    for i in range(2):
-        with open('/proc/stat') as f:
-            fields = [float(column) for column in f.readline().strip().split()[1:]]
-        idle, total = fields[3], sum(fields)
-        idle_delta, total_delta = idle - last_idle, total - last_total
-        last_idle, last_total = idle, total
-        if i == 0:
-            exit_event.wait(interval)
-
-    return 100.0 * (1.0 - idle_delta / total_delta)
-
-
 def get_value_for_bits(val, from_bit=0, to_bit=63):
     mask = sum(2 ** x for x in range(from_bit, to_bit + 1))
     return (val & mask) >> from_bit
@@ -302,7 +311,7 @@ def undervolt(config):
             read_value = get_undervolt(plane)[plane]
             read_offset_mv = calc_undervolt_mv(read_value)
             match = OK if write_value == read_value else ERR
-            print(
+            log(
                 '[D] Undervolt plane {:s} - write {:.0f} mV ({:#x}) - read {:.0f} mV ({:#x}) - match {}'.format(
                     plane, write_offset_mv, write_value, read_offset_mv, read_value, match
                 )
@@ -350,7 +359,7 @@ def set_icc_max(config):
                     read_value = get_icc_max(plane)[plane]
                     read_current_A = calc_icc_max_amp(read_value)
                     match = OK if write_value == read_value else ERR
-                    print(
+                    log(
                         '[D] IccMax plane {:s} - write {:.2f} A ({:#x}) - read {:.2f} A ({:#x}) - match {}'.format(
                             plane, write_current_amp, write_value, read_current_A, read_value, match
                         )
@@ -368,7 +377,7 @@ def load_config():
         for option in ('Update_Rate_s', 'PL1_Tdp_W', 'PL1_Duration_s', 'PL2_Tdp_W', 'PL2_Duration_S'):
             value = config.getfloat(power_source, option, fallback=None)
             if value is not None:
-                value = config.set(power_source, option, str(max(0.1, value)))
+                value = config.set(power_source, option, str(max(0.001, value)))
             elif option == 'Update_Rate_s':
                 fatal('The mandatory "Update_Rate_s" parameter is missing.')
 
@@ -377,7 +386,7 @@ def load_config():
             valid_trip_temp = min(TRIP_TEMP_RANGE[1], max(TRIP_TEMP_RANGE[0], trip_temp))
             if trip_temp != valid_trip_temp:
                 config.set(power_source, 'Trip_Temp_C', str(valid_trip_temp))
-                print(
+                log(
                     '[!] Overriding invalid "Trip_Temp_C" value in "{:s}": {:.1f} -> {:.1f}'.format(
                         power_source, trip_temp, valid_trip_temp
                     )
@@ -391,7 +400,7 @@ def load_config():
                 valid_value = min(0, value)
                 if value != valid_value:
                     config.set(key, plane, str(valid_value))
-                    print(
+                    log(
                         '[!] Overriding invalid "{:s}" value in "{:s}" voltage plane: {:.0f} -> {:.0f}'.format(
                             key, plane, value, valid_value
                         )
@@ -407,6 +416,13 @@ def load_config():
                 value = config.getfloat(key, plane, fallback=0.0)
                 config.set(key, plane, str(value))
 
+    # Check for CORE/CACHE values mismatch
+    for key in UNDERVOLT_KEYS:
+        if key in config:
+            if config.getfloat(key, 'CORE', fallback=0) != config.getfloat(key, 'CACHE', fallback=0):
+                warning('On Skylake and newer CPUs CORE and CACHE values should match!')
+                break
+
     iccmax_enabled = False
     # check for invalid values (ie. <= 0 or > 0x3FF) in the IccMax settings
     for key in ICCMAX_KEYS:
@@ -418,7 +434,7 @@ def load_config():
                         raise ValueError
                     iccmax_enabled = True
                 except ValueError:
-                    warning('Invalid value for {:s} in {:s}'.format(plane, key))
+                    warning('Invalid value for {:s} in {:s}'.format(plane, key), oneshot=False)
                     config.remove_option(key, plane)
                 except configparser.NoOptionError:
                     pass
@@ -445,7 +461,7 @@ def calc_reg_values(platform_info, config):
                 trip_offset = int(round(critical_temp - Trip_Temp_C))
                 regs[power_source]['MSR_TEMPERATURE_TARGET'] = trip_offset << 24
             else:
-                print('[I] {:s} trip temperature is disabled in config.'.format(power_source))
+                log('[I] {:s} trip temperature is disabled in config.'.format(power_source))
 
         power_unit = get_power_unit()
 
@@ -458,78 +474,89 @@ def calc_reg_values(platform_info, config):
             cur_pkg_power_limits = get_cur_pkg_power_limits()
             if PL1_Tdp_W is None:
                 PL1 = cur_pkg_power_limits['PL1']
-                print('[I] {:s} PL1_Tdp_W disabled in config.'.format(power_source))
+                log('[I] {:s} PL1_Tdp_W disabled in config.'.format(power_source))
             else:
                 PL1 = int(round(PL1_Tdp_W / power_unit))
 
             if PL1_Duration_s is None:
                 TW1 = cur_pkg_power_limits['TW1']
-                print('[I] {:s} PL1_Duration_s disabled in config.'.format(power_source))
+                log('[I] {:s} PL1_Duration_s disabled in config.'.format(power_source))
             else:
                 Y, Z = calc_time_window_vars(PL1_Duration_s)
                 TW1 = Y | (Z << 5)
 
             if PL2_Tdp_W is None:
                 PL2 = cur_pkg_power_limits['PL2']
-                print('[I] {:s} PL2_Tdp_W disabled in config.'.format(power_source))
+                log('[I] {:s} PL2_Tdp_W disabled in config.'.format(power_source))
             else:
                 PL2 = int(round(PL2_Tdp_W / power_unit))
 
             if PL2_Duration_s is None:
                 TW2 = cur_pkg_power_limits['TW2']
-                print('[I] {:s} PL2_Duration_s disabled in config.'.format(power_source))
+                log('[I] {:s} PL2_Duration_s disabled in config.'.format(power_source))
             else:
                 Y, Z = calc_time_window_vars(PL2_Duration_s)
                 TW2 = Y | (Z << 5)
 
             regs[power_source]['MSR_PKG_POWER_LIMIT'] = (
-                PL1 | (1 << 15) | (TW1 << 17) | (PL2 << 32) | (1 << 47) | (TW2 << 49)
+                PL1 | (1 << 15) | (1 << 16) | (TW1 << 17) | (PL2 << 32) | (1 << 47) | (TW2 << 49)
             )
         else:
-            print('[I] {:s} package power limits are disabled in config.'.format(power_source))
+            log('[I] {:s} package power limits are disabled in config.'.format(power_source))
 
         # cTDP
         c_tdp_target_value = config.getint(power_source, 'cTDP', fallback=None)
         if c_tdp_target_value is not None:
             if platform_info['feature_programmable_tdp_limit'] != 1:
-                print("[W] cTDP setting not supported by this CPU")
+                log("[W] cTDP setting not supported by this CPU")
             elif platform_info['number_of_additional_tdp_profiles'] < c_tdp_target_value:
-                print("[W] the configured cTDP profile is not supported by this CPU")
+                log("[W] the configured cTDP profile is not supported by this CPU")
             else:
                 valid_c_tdp_target_value = max(0, c_tdp_target_value)
                 regs[power_source]['MSR_CONFIG_TDP_CONTROL'] = valid_c_tdp_target_value
     return regs
 
 
-def set_hwp(pref):
-    # set HWP energy performance hints
-    assert pref in ('performance', 'balance_performance', 'default', 'balance_power', 'power')
-    CPUs = [
-        '/sys/devices/system/cpu/cpu{:d}/cpufreq/energy_performance_preference'.format(x) for x in range(cpu_count())
-    ]
-    for i, c in enumerate(CPUs):
-        with open(c, 'wb') as f:
-            f.write(pref.encode())
-        if args.debug:
-            with open(c) as f:
-                read_value = f.read().strip()
-                match = OK if pref == read_value else ERR
-                print('[D] HWP for cpu{:d} - write "{:s}" - read "{:s}" - match {}'.format(i, pref, read_value, match))
+def set_hwp():
+    # set HWP energy performance preference
+    cur_val = readmsr(0x774, cpu=0)
+    new_val = (cur_val & 0xFFFFFFFF00FFFFFF) | (HWP_VALUE << 24)
+
+    writemsr(0x774, new_val)
+    if args.debug:
+        read_value = readmsr(0x774, from_bit=24, to_bit=31)[0]
+        match = OK if HWP_VALUE == read_value else ERR
+        log('[D] HWP - write "{:#02x}" - read "{:#02x}" - match {}'.format(HWP_VALUE, read_value, match))
+
+
+def set_disable_bdprochot():
+    # Disable BDPROCHOT
+    cur_val = readmsr(0x1FC, flatten=True)
+    new_val = cur_val & 0xFFFFFFFFFFFFFFFE
+
+    writemsr(0x1FC, new_val)
+    if args.debug:
+        read_value = readmsr(0x1FC, from_bit=31, to_bit=31)[0]
+        match = OK if ~read_value else ERR
+        log('[D] BDPROCHOT - write "{:#02x}" - read "{:#02x}" - match {}'.format(0, read_value, match))
 
 
 def power_thread(config, regs, exit_event):
     try:
         mchbar_mmio = MMIO(0xFED159A0, 8)
     except MMIOError:
-        fatal('Unable to open /dev/mem. Try to disable Secure Boot.')
+        warning('Unable to open /dev/mem. TDP override might not work correctly.')
+        warning('Try to disable Secure Boot and/or enable CONFIG_DEVMEM in kernel config.')
+        mchbar_mmio = None
 
+    next_hwp_write = 0
     while not exit_event.is_set():
-        # print thermal status
+        # log thermal status
         if args.debug:
             thermal_status = get_reset_thermal_status()
             for index, core_thermal_status in enumerate(thermal_status):
                 for key, value in core_thermal_status.items():
-                    print('[D] core {} thermal status: {} = {}'.format(index, key.replace("_", " "), value))
+                    log('[D] core {} thermal status: {} = {}'.format(index, key.replace("_", " "), value))
 
         # switch back to sysfs polling
         if power['method'] == 'polling':
@@ -542,7 +569,7 @@ def power_thread(config, regs, exit_event):
             if args.debug:
                 read_value = readmsr(0x1A2, 24, 29, flatten=True)
                 match = OK if write_value >> 24 == read_value else ERR
-                print(
+                log(
                     '[D] TEMPERATURE_TARGET - write {:#x} - read {:#x} - match {}'.format(
                         write_value >> 24, read_value, match
                     )
@@ -555,7 +582,7 @@ def power_thread(config, regs, exit_event):
             if args.debug:
                 read_value = readmsr(0x64B, 0, 1, flatten=True)
                 match = OK if write_value == read_value else ERR
-                print(
+                log(
                     '[D] CONFIG_TDP_CONTROL - write {:#x} - read {:#x} - match {}'.format(
                         write_value, read_value, match
                     )
@@ -567,32 +594,43 @@ def power_thread(config, regs, exit_event):
         if args.debug:
             read_value = readmsr(0x610, 0, 55, flatten=True)
             match = OK if write_value == read_value else ERR
-            print(
+            log(
                 '[D] MSR PACKAGE_POWER_LIMIT - write {:#x} - read {:#x} - match {}'.format(
                     write_value, read_value, match
                 )
             )
-        # set MCHBAR register to the same PL1/2 values
-        mchbar_mmio.write32(0, write_value & 0xFFFFFFFF)
-        mchbar_mmio.write32(4, write_value >> 32)
-        if args.debug:
-            read_value = mchbar_mmio.read32(0) | (mchbar_mmio.read32(4) << 32)
-            match = OK if write_value == read_value else ERR
-            print(
-                '[D] MCHBAR PACKAGE_POWER_LIMIT - write {:#x} - read {:#x} - match {}'.format(
-                    write_value, read_value, match
+        if mchbar_mmio is not None:
+            # set MCHBAR register to the same PL1/2 values
+            mchbar_mmio.write32(0, write_value & 0xFFFFFFFF)
+            mchbar_mmio.write32(4, write_value >> 32)
+            if args.debug:
+                read_value = mchbar_mmio.read32(0) | (mchbar_mmio.read32(4) << 32)
+                match = OK if write_value == read_value else ERR
+                log(
+                    '[D] MCHBAR PACKAGE_POWER_LIMIT - write {:#x} - read {:#x} - match {}'.format(
+                        write_value, read_value, match
+                    )
                 )
-            )
+
+        # Disable BDPROCHOT
+        disable_bdprochot = config.getboolean(power['source'], 'Disable_BDPROCHOT', fallback=None)
+        if disable_bdprochot is not None:
+            set_disable_bdprochot()
 
         wait_t = config.getfloat(power['source'], 'Update_Rate_s')
         enable_hwp_mode = config.getboolean('AC', 'HWP_Mode', fallback=False)
-        if power['source'] == 'AC' and enable_hwp_mode:
-            cpu_usage = cpu_usage_pct(exit_event, interval=wait_t)
-            # set full performance mode only when load is greater than this threshold (~ at least 1 core full speed)
-            performance_mode = cpu_usage > 100.0 / (cpu_count() * 1.25)
-            # check again if we are on AC, since in the meantime we might have switched to BATTERY
-            if not is_on_battery(config):
-                set_hwp('performance' if performance_mode else 'balance_performance')
+        # set HWP less frequently. Just to be safe since (e.g.) TLP might reset this value
+        if (
+            enable_hwp_mode
+            and next_hwp_write <= time()
+            and (
+                (power['method'] == 'dbus' and power['source'] == 'AC')
+                or (power['method'] == 'polling' and not is_on_battery(config))
+            )
+        ):
+            set_hwp()
+            next_hwp_write = time() + HWP_INTERVAL
+
         else:
             exit_event.wait(wait_t)
 
@@ -615,10 +653,11 @@ def check_kernel():
         except (subprocess.CalledProcessError, IOError):
             pass
     if kernel_config is None:
-        print('[W] Unable to obtain and validate kernel config.')
+        log('[W] Unable to obtain and validate kernel config.')
+        return
     elif not re.search('CONFIG_DEVMEM=y', kernel_config):
-        fatal('Bad kernel config: you need CONFIG_DEVMEM=y.')
-    elif not re.search('CONFIG_X86_MSR=(y|m)', kernel_config):
+        warning('Bad kernel config: you need CONFIG_DEVMEM=y.')
+    if not re.search('CONFIG_X86_MSR=(y|m)', kernel_config):
         fatal('Bad kernel config: you need CONFIG_X86_MSR builtin or as module.')
 
 
@@ -648,7 +687,7 @@ def check_cpu():
         if cpuinfo['cpu family'] != 6 or cpu_model is None:
             fatal('Your CPU model is not supported.')
 
-        print('[I] Detected CPU architecture: Intel {:s}'.format(cpu_model))
+        log('[I] Detected CPU architecture: Intel {:s}'.format(cpu_model))
     except:
         fatal('Unable to identify CPU model.')
 
@@ -676,13 +715,13 @@ def monitor(exit_event, wait):
 
     undervolt_values = get_undervolt(convert=True)
     undervolt_output = ' | '.join('{:s}: {:.2f} mV'.format(plane, undervolt_values[plane]) for plane in VOLTAGE_PLANES)
-    print('[D] Undervolt offsets: {:s}'.format(undervolt_output))
+    log('[D] Undervolt offsets: {:s}'.format(undervolt_output))
 
     iccmax_values = get_icc_max(convert=True)
     iccmax_output = ' | '.join('{:s}: {:.2f} A'.format(plane, iccmax_values[plane]) for plane in CURRENT_PLANES)
-    print('[D] IccMax: {:s}'.format(iccmax_output))
+    log('[D] IccMax: {:s}'.format(iccmax_output))
 
-    print('[D] Realtime monitoring of throttling causes:\n')
+    log('[D] Realtime monitoring of throttling causes:\n')
     while not exit_event.is_set():
         value = readmsr(IA32_THERM_STATUS, from_bit=0, to_bit=15, cpu=0)
         offsets = {'Thermal': 0, 'Power': 10, 'Current': 12, 'Cross-domain (e.g. GPU)': 14}
@@ -700,7 +739,11 @@ def monitor(exit_event, wait):
             )
             stats2[power_plane] = '{:.1f} W'.format(energy_w)
         output2 = ('{:s}: {:s}'.format(label, stats2[label]) for label in stats2)
-        print('[{}] {}  ||  {}{}'.format(power['source'], ' - '.join(output), ' - '.join(output2), ' ' * 10), end='\r')
+        terminator = '\n' if args.log else '\r'
+        log(
+            '[{}] {}  ||  {}{}'.format(power['source'], ' - '.join(output), ' - '.join(output2), ' ' * 10),
+            end=terminator,
+        )
         exit_event.wait(wait)
 
 
@@ -720,23 +763,32 @@ def main():
     )
     parser.add_argument('--config', default='/etc/lenovo_fix.conf', help='override default config file path')
     parser.add_argument('--force', action='store_true', help='bypass compatibility checks (EXPERTS only)')
+    parser.add_argument('--log', metavar='/path/to/file', help='log to file instead of stdout')
     args = parser.parse_args()
 
+    if args.log:
+        try:
+            args.log = open(args.log, 'w')
+        except:
+            args.log = None
+            fatal('Unable to write to the log file!')
+
     if not args.force:
         check_kernel()
         check_cpu()
 
-    print('[I] Loading config file.')
+    log('[I] Loading config file.')
     config = load_config()
     power['source'] = 'BATTERY' if is_on_battery(config) else 'AC'
 
     platform_info = get_cpu_platform_info()
     if args.debug:
         for key, value in platform_info.items():
-            print('[D] cpu platform info: {} = {}'.format(key.replace("_", " "), value))
+            log('[D] cpu platform info: {} = {}'.format(key.replace("_", " "), value))
     regs = calc_reg_values(platform_info, config)
 
     if not config.getboolean('GENERAL', 'Enabled'):
+        log('[I] Throttled is disabled in config file... Quitting. :(')
         return
 
     exit_event = Event()
@@ -777,7 +829,7 @@ def main():
         path="/org/freedesktop/UPower/devices/line_power_AC",
     )
 
-    print('[I] Starting main loop.')
+    log('[I] Starting main loop.')
 
     if args.monitor is not None:
         monitor_thread = Thread(target=monitor, args=(exit_event, args.monitor))
